#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEServer.h>
#include <BLE2902.h>
#include "e3bot-actuators.hpp"
#include "util.hpp"

// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define SERVICE_UUID        "ecf085e8-a276-43f5-8247-2d8d818d9d31"
#define CHARACTERISTIC_UUID "50fa11b2-4bdb-49a5-85a4-9835f06d47f8"
#define DEVICE_NAME         "COG21"
#define VERSION             "1.0.0"

const uint8_t PIN_SERVO_WHEEL_L = 12; // Dev: 14
const uint8_t PIN_SERVO_WHEEL_R = 26; // Dev: 27
const uint8_t PIN_SERVO_HEAD_L  = 14; // Dev: 26
const uint8_t PIN_SERVO_HEAD_R  = 02; // Dev: 25

const unsigned int TIMEOUT_MS = 5000;
const unsigned int EVENT_CONNECTED = 1 << 0;
const unsigned int EVENT_DISCONNECTED = 1 << 1;
const uint16_t CONNECTED_NONE = -1;
unsigned int events = 0;
unsigned int timefromlastmessage = 0;
uint16_t connectedTo = CONNECTED_NONE;
bool isAdvertising = false;
BLE2902 descClientChar;

e3bot::Actuators actuators({
    .pinWheelL = PIN_SERVO_WHEEL_L,
    .pinWheelR = PIN_SERVO_WHEEL_R,
    .pinHeadL  = PIN_SERVO_HEAD_L,
    .pinHeadR  = PIN_SERVO_HEAD_R
});

namespace e3bot
{
  void setAdvertisingState(bool newState) {
    isAdvertising = newState;
    BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
    if (newState) {
      pAdvertising->start();
    } else {
      pAdvertising->stop();
    }
  }
  
  class BotServerCallbacks: public BLEServerCallbacks {
      void onConnect(BLEServer* pServer) {
        auto peerDevices = pServer->getPeerDevices(false);
        
        if (connectedTo != CONNECTED_NONE) {
          // Abort - we are already connected
          
          for (auto pair : peerDevices) {
            if (pair.first != connectedTo) {
              pServer->disconnect(pair.first);
              pServer->removePeerDevice(pair.first, false);
            }
          }

          Serial.println("onConnect rejected");
          return;
        }

        
        for (auto pair : peerDevices) {
          connectedTo = pair.first;
        }
        
        Serial.println("onConnect accepted");
        timefromlastmessage = 0;
        events |= EVENT_CONNECTED;
        setAdvertisingState(false);
      };
      
      void onDisconnect(BLEServer* pServer) {
        if (connectedTo == CONNECTED_NONE) { return; }

        auto peerDevices = pServer->getPeerDevices(false); 
        if (peerDevices.find(connectedTo) != peerDevices.end()) { return; }
        
        Serial.println("onDisconnect accepted");
        events |= EVENT_DISCONNECTED;
        connectedTo = CONNECTED_NONE;
        setAdvertisingState(true);
      }
  };
  
  class BotCharacteristicCallbacks : public BLECharacteristicCallbacks
  {
  public:
    void onRead(BLECharacteristic *pCharacteristic) {
      // Serial.println("onRead");
    }
    void onWrite(BLECharacteristic *pCharacteristic) {
      std::string value(pCharacteristic->getValue());
      // Serial.println("onWrite " + String(value.c_str()));
      timefromlastmessage = 0;

      if (value.size() < 1) { return; }
      switch (value[0])
      {
      case 'V':
        delay(5);
        pCharacteristic->setValue("V" + std::string(VERSION));
        pCharacteristic->notify();
        break;
      case 'M':
        {
          struct OpM {
            int8_t inByteFwd;
            int8_t inByteTurn;
            int8_t inByteHead;
            int8_t inByteNServos;
          };

          int8_t valueSize = value.size();
          if (valueSize >= 1 + sizeof(OpM)) {
            OpM &opM = *reinterpret_cast<OpM*>(&value[1]);
            double inputFwd  = e3bot::clamp(opM.inByteFwd  / 127.0, -1.0, 1.0);
            double inputTurn = e3bot::clamp(opM.inByteTurn / 127.0, -1.0, 1.0);
            double inputHead = e3bot::clamp(opM.inByteHead / 127.0, -1.0, 1.0);

            actuators.setInput(inputFwd, inputTurn, inputHead);

            char *servoData = &value[1 + sizeof(OpM)];
            for (int8_t i = 0; i < opM.inByteNServos; ++i) {
              int8_t servoNum = (int8_t)*(servoData++);
              double servoVal = e3bot::clamp((int8_t)*(servoData++) / 127.0, -1.0, 1.0);
              
              actuators.setServo(servoNum, servoVal);
            }
          }
        }
        break;
      case 'R':
        actuators.reset();
        
        break;
      }
    }
    void onNotify(BLECharacteristic* pCharacteristic) {
      // Serial.println("onNotify");
    }
    void onStatus(BLECharacteristic* pCharacteristic, Status s, uint32_t code) {
      Serial.println("onStatus " + String(s) + "/" + String(code));
    }
  };
}

void setup() {
  Serial.begin(115200);
  //Serial.println("Starting BLE");

  BLEDevice::init(DEVICE_NAME);
  BLEServer *pServer = BLEDevice::createServer();
  pServer->setCallbacks(new e3bot::BotServerCallbacks());
  
  BLEService *pService = pServer->createService(SERVICE_UUID);
  e3bot::BotCharacteristicCallbacks *pHandler = new e3bot::BotCharacteristicCallbacks();

  BLECharacteristic *pCharacteristic = pService->createCharacteristic(
    CHARACTERISTIC_UUID,
    BLECharacteristic::PROPERTY_READ |
    BLECharacteristic::PROPERTY_WRITE |
    BLECharacteristic::PROPERTY_NOTIFY |
    BLECharacteristic::PROPERTY_INDICATE
  );

  descClientChar.setNotifications(true);
  pCharacteristic->addDescriptor(&descClientChar);
  pCharacteristic->setCallbacks(pHandler);
  
  pService->start();
  BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
  pAdvertising->addServiceUUID(SERVICE_UUID);
  pAdvertising->setScanResponse(true);
  pAdvertising->setMinPreferred(0x06);
  pAdvertising->setMinPreferred(0x12);
  e3bot::setAdvertisingState(true);
  Serial.println("Characteristic defined");
}

void resetBot() {
  actuators.reset();
  connectedTo = CONNECTED_NONE;
  timefromlastmessage = 0;
  delayFor(500);
  e3bot::setAdvertisingState(true);
  Serial.println("Resetting");
}

void delayFor(unsigned long ms) {
  actuators.updateServos(ms);
  if (connectedTo != CONNECTED_NONE) {
    timefromlastmessage += ms;
    if (timefromlastmessage > TIMEOUT_MS) {
      resetBot();
    }
  }
  delay(ms);
}

void loop() {
  if (events == 0) {
    delayFor(20);
  } else {
    // Serial.println("Got event!");
    if (events & EVENT_CONNECTED) {
      Serial.println("EVENT_CONNECTED");
      actuators.setInput(0.0, 0.0, 0.0);
    }
    if (events & EVENT_DISCONNECTED) {
      Serial.println("EVENT_DISCONNECTED");
      resetBot();
    }
    
    events = 0;
  }
}
